#! /usr/bin/perl

# POT by Steven Price

#$debug=1;
#$debugpause=0;
#$debugscore=1;

use Socket;

# Use some default values if none are given
$remote = shift || 'localhost';
$port   = shift || 20005;

# The boring tcp stuff
$iaddr = inet_aton($remote) or die "no host: $remote";
$paddr = sockaddr_in($port,$iaddr);
$proto = getprotobyname('tcp');
socket(SOCK, PF_INET, SOCK_STREAM, $proto) or die "socket: $!";
connect(SOCK,$paddr) or die "connect: $!";

print "Connected!\n" if $debug;

# We should be connected - on to the fun stuff
send(SOCK,"Player\n",0);

getline() =~ /(\d+) (\d+)/;
($width,$height)=($1,$2);
print "${width}x$height\n" if $debug;
for($a=1;$a<=$height;$a++) {
    $board[$a]=$l=getline();
    $ox=0;
    while ($l=~/\@/) {
	$homey=$a;
	$ox=$homex=index($l,"@")+1+$ox;
	$l=$'; #'
	push @homex,$homex;
	push @homey,$homey;
    }
}

push @homex,-1;
push @homey,-1;

$homex=shift(@homex);$homey=shift(@homey);
push @homex,$homex;push @homey,$homey;

@rmx=(1,-1,0,0,1,-1,0,0);
@rmy=(0,0,1,-1,0,0,1,-1);
@rm=('E','W','N','S','E','W','N','S');

getline() =~ /(\d+) (\d+) (\d+)/;
($us,$maxweight,$money)=($1,$2,$3);

print "Home at $homex,$homey\n" if $debug;

$norand=1;

if ($debug) {
    $scrnwidth=($ENV{'COLUMNS'} || 150);
}

# The main loop
while (1) {
    print "Reading playerdata...\n" if $debug;
    $playerdata=getline() or die "Dead";
    # Remove all players from list
    undef %players;
    $playercount=0;
    print $playerdata if $debug;
    while ($playerdata =~ /\#(\d+)\s*([^\#]*)/)
    {
	$playerid=$1;
	$playercommands=$2;
	$playerdata=$'; # Silly emacs cant cope with $' - hence this comment
	$playercount++;

	# Add the player back into the list
	$players{$playerid}=1;
	while ($playercommands =~ /^\s*(\S)/) {
	    if ($1 eq 'N') {
		$playery{$playerid}++;
		$playercommands=$'; #'
	    } elsif ($1 eq 'S') {
		$playery{$playerid}--;
		$playercommands=$'; #'
	    } elsif ($1 eq 'E') {
		$playerx{$playerid}++;
		$playercommands=$'; #'
	    } elsif ($1 eq 'W') {
		$playerx{$playerid}--;
		$playercommands=$'; #'
	    } elsif ($1 eq 'P') {
		$playercommands =~ /^\s*. (\d+)/;
		$playercommands = $'; # '
		$packid=$1;
		$packx{$packid}=$playerid;
		$packy{$packid}=0;
		if ($playerid==$us) {
		    $ourpacks{$packid}=1;
		}
	    } elsif ($1 eq 'D') {
		$playercommands =~ /^\s*. (\d+)/;
		$playercommands = $'; #'
		$packid=$1;
		$packx{$packid}=$playerx{$playerid};
		$packy{$packid}=$playery{$playerid};
		if ($playerid==$us) {
		    delete $ourpacks{$packid};
		    if ($packdestx{$packid}==$playerx{$us} &&
			$packdesty{$packid}==$playery{$us}) {
			$score+=$packweight{$packid};
		    }
		}
		if ($packdestx{$packid}==$playerx{$playerid} &&
		    $packdesty{$packid}==$playery{$playerid}) {
		    # Package has been dropped on destination - so remove it
		    delete $packx{$packid};
		}
	    } elsif ($1 eq 'X') {
		$playercommands =~ /^\s*. (\d+)/;
                $playercommands = $'; #'
		$playerx{$playerid}=$1;
	    } elsif ($1 eq 'Y') {
		$playercommands =~ /^\s*\S\s+(\d+)/;
                $playercommands = $'; #'
		$playery{$playerid}=$1;
	    } else {
		die "Didn't understand server response!";
	    }
	}# playercommands
    }# playerdata
    $packlist=getline() or die "Dead";

    # Remove all the packages on our square - so we'll know if packages have
    # been deleted
    foreach(keys %packx) {
	if ($packx{$_}==$playerx{$us} && $packy{$_}==$playery{$us}) {
	    delete $packx{$_};
	}
    }

    undef @packhere;
    while ($packlist =~ /^(\d+) (\d+) (\d+) (\d+)\s*/) {
	$packid=$1;
	push @packhere,$packid;
	$packdestx{$packid}=$2;
	$packdesty{$packid}=$3;
	$packweight{$packid}=$4;
	$packx{$packid}=$playerx{$us};
	$packy{$packid}=$playery{$us};
	$packlist=$'; #'
    }

    $curw=0;
    $hpcount=0;
    foreach(keys %ourpacks) {
        $curw += $packweight{$_};
	$hpcount++;
    }

    $sparew=$maxweight-$curw;

    $pickup='';

    # See if we can pick up any packs on the ground...
    # Sort the list so that the largest weights come first
    @opackhere = sort {$packweight{$b} <=> $packweight{$a}} @packhere;

    # First we'll only consider packages that have the same destination as a
    # package we already hold
    undef $topickup;
    foreach(@opackhere) {
	if (!$givenup{$_} && $packweight{$_}<=$sparew) {
	    $pid=$_;
	    $ok=2;
	    foreach(keys %ourpacks) {
		$ok=0 if ($ok==2);
		if ($packdestx{$_}==$packdestx{$pid} &&
		    $packdesty{$_}==$packdesty{$pid}) {
		    $ok=1;
		}
	    }
	    if (!$ok) {
		foreach(keys %topickup) {
		    if ($packdestx{$_}==$packdestx{$pid} &&
			$packdesty{$_}==$packdesty{$pid}) {
			$ok=1;
		    }
		}
	    }
	    if ($ok) {
		$pickup .= "$pid ";
		$topickup{$pid}=1;
		$sparew-=$packweight{$pid};
	    }
	}
    }

    # Then we add any that are left and can fit
    foreach(@opackhere) {
        if (!$givenup{$_} && !$topickup{$_} && $packweight{$_}<=$sparew) {
            $pickup .= "$_ ";
	    $topickup{$_}=1;
            $sparew-=$packweight{$_};
        }
    }

    # Work out how much we're going to spend...
    $spend=$money/100; # If we've got loads of money we might as well waste it!
    $spend=int($spend);
    $spend=1 if ($spend<1); # We're not going to use negative numbers because
                            # they would be a pain! :)

    if ($playercount==1) {
	# If we're the only player we don't need to waste our money
	$spend=1;
    }

    $oput="Pick";# Dummy output in case it all goes pearshaped

    # If we're going to be picking up packs - then thats the end of our turn
    if ($pickup) {
	chomp $pickup; # Trim the trailing space in case that gets us killed
	$oput="Pick $pickup";
	undef %nogo;
    } else {
	# We're not picking up - so on to working out if we need to drop
	$drop='';
	$dropvalue=0;
	foreach(keys %ourpacks) {
	    if ($packdestx{$_}==$playerx{$us} &&
		$packdesty{$_}==$playery{$us}) {
		$drop .= "$_ ";
		$dropvalue+=$packweight{$_};
	    }
	}
	if ($drop) {
	    chomp $drop;
	    $oput="Drop $drop";
	    print "Dropping $dropvalue of packages\n" if ($debugscore);
	} else {
	    # We haven't got anything to pickup or drop - so we've got to work
	    # out where we're going to move to
	    $mdist=$width+$height;
	    $x=$playerx{$us};$y=$playery{$us};
	    # Ensure that we don't change the package we're trying to deliver
	    # mid-delivery
	    if (!$ourpacks{$curpack}) {
		$curpack=-1;
		foreach (keys %ourpacks) {
		    $pid=$_;
		    if (!$givenup{$pid}) {
			$dist=abs($x-$packdestx{$pid})+
			    abs($y-$packdesty{$pid});
			if ($dist<$mdist) {
			    $curpack=$pid;
			    $mdist=$dist;
			}
		    }
		}
	    }
	    if ($curpack!=-1) {
		# We've got a package to deliver
		$dx=$packdestx{$curpack};
		$dy=$packdesty{$curpack};
		print "Trying to deliver $curpack to $dx,$dy\n" if $debug;

		$dir=FindRoute($dx,$dy,$x,$y);
		if ($dir) {
		    $oput="Move $dir";
		} else {
		    # Can't move towards destination - we give up on this
		    # package
		    $givenup{$curpack}=1;
		    $oput="Drop $curpack";
		    print "Given up on package $curpack\n" if $debug;
		}
		$maxrandbit=0;
	    } elsif ($curpack==-1) {
		$dir = FindRoute($homex,$homey,$x,$y);
		if (abs($homex-$x)+abs($homey-$y)<5) {
		    $nearhome++;
		    if ($nearhome>20+rand()*10) {
			# We've been near the home for a while and not got
			# there - its probably a stalemate with another bot
			# So we give in and choose another home to go to (the
			# rand()*10 bit should stop us both choosing the other
			# home at the same time)
			$homex=shift(@homex);$homey=shift(@homey);
			push @homex,$homex;push @homey,$homey;
		    }
		} else {
		    $nearhome=0;
		}
		if ($dir) {
		    $oput = "Move $dir";
		} else {
		    # We can't find any more moves to get to the home
		    # so either its a badly designed map - with an impossible
		    # to reach home - or we're standing on it - either way
		    # lets try the next home...
		    $homex=shift(@homex);$homey=shift(@homey);
		    push @homex,$homex;push @homey,$homey;
		    # But since we're really lazy (and to stop infinite loops)
		    # we're just going to sit here for a turn
		    $oput = "Pick";
		}
		if ($homex==-1) {
		    # This is a special homex thing which we replace with
		    # the location of a package (if there is one lying around
		    # somewhere)
		    $mdist=$width+$height;
		    foreach(keys %packx) {
			if ($packy{$_}) {
			    $dist=abs($packx{$_}-$x)+abs($packy{$_}-$y);
			    if ($dist<$mdist) {
				$homex=$packx{$_};
				$homey=$packy{$_};
				$mdist=$dist;
			    }
			}
		    }
		    if ($homex==-1) {
			# Couldn't find a package lying around - go back to
			# looking on the home squares
			$homex=shift(@homex);$homey=shift(@homey);
			push @homex,$homex;push @homey,$homey;
		    }
		}
	    } # curpack==-1
	} # (!$drop)
    } # (!$pickup)

    $spend=1; # We're not going to bother with different spendings :)
    # Really because we're not bothering to look at what other robots are doing
    # there isn't any point spending any more than 1 - because we don't have
    # any idea whether we need to be conserned with the order of execution of
    # commands (and it would be difficult to tell if we want our command
    # executed before or after the other robots). So we're just going to stick
    # with 1 because it gives us the maximum number of turns!

    send(SOCK,"$spend $oput\n",0);

    # Debugging code moved until after we send move - so we don't leave the
    # server waiting for us for as long :)
    if ($debug) {
        # Nice debugging code - we print out what we think is the current
        # board state...
        $cellwidth=int($scrnwidth/$width);
        for($a=$height;$a>=1;$a--) {
            $olen=0;
            @packsonline=();
            $l='';
            foreach(keys %packx) {
                if ($packy{$_}==$a) {
                    push @packsonline,$_;
                }
            }
            for($b=1;$b<=$width;$b++) {
                $o='';
                foreach(keys %players) {
                    if ($playerx{$_}==$b && $playery{$_}==$a) {
                        $o.=$_;
                    }
                }
                $pcount=0;
                foreach(@packsonline) {
                    if ($packx{$_}==$b) {
                        $pcount++;
                    }
                }
                $o.="X" if ($nogo{"$b,$a"});
                $o.=substr($board[$a],$b-1,1);
                if ($pcount) {$o.="P".$pcount;}
                $len=length($o);
                if ($len>$cellwidth) {$olen+=$len-$cellwidth;}
                if ($olen) {
                    $olen-- if ($len<$cellwidth);
                    $l.=sprintf "%".($cellwidth-1)."s",$o;
                } else {
                    $l.=sprintf "%".$cellwidth."s",$o;
                }
            }
            print "$l\n";
        }
        print "We're $us at ".$playerx{$us}.",".$playery{$us};
        print " holding $hpcount ";
        print "packages weighing $curw/$maxweight\n";
        # We only flush here because then when its going a bit slow ('cos
        # there's lots of packages normally) it'll only display when its
        # all done
        $|=1;
        <STDIN> if $debugpause;
        $|=0;
    }

    print "$spend $oput\n" if $debug;

    $money-=$spend;
    print "Money left: $money, Score: $score\n" if ($debugscore);
    if ($money<0) {
	print "We should die now!\n" if $debug;
    }
}# while 1


sub getline
{
    my($a);
    # I'm sure there must be a better way
    $a='';
    while ($a !~ /\n/) {
	if (!defined(recv(SOCK,$b,1,0))) {die "Recv failed";}
	if ($b eq "") {die "No data read '$a' in buffer";}
	$a.=$b;
    }
    return $a;
}

sub CalcRoute
{
    my ($x,$y)=@_;

    my (@todox,@todoy);
    my @map;

    my $a,$b,$nx,$ny,$tx,$ty;

    @map=@board;
    push @todox,$x;
    push @todoy,$y;

    # Mark the destination as special
    substr($map[$y],$x-1,1)="D";

    while (($tx=shift(@todox)) && ($ty=shift(@todoy))) {
	# Randomise so that this bot can play against itself better
	$b=int(rand()*4);
	for($a=$b;$a<$b+4;$a++) {
	    $nx=$tx-$rmx[$a];
	    $ny=$ty-$rmy[$a];
	    if ($nx>=1 && $nx<=$width &&
		$ny>=1 && $ny<=$height) {
		$t=substr($map[$ny],$nx-1,1);
		if ($t eq '.' || $t eq '@') {
		    substr($map[$ny],$nx-1,1)=$rm[$a];
		    push @todox,$nx;
		    push @todoy,$ny;
		}
	    }
	}
	$count=@todox;
	#print "$count\n";
    }

    $route{"$x,$y"}=\@map;
}

sub FindRoute
{
    my ($dx,$dy,$cx,$cy)=@_;
    my $a;

    if (!$route{"$dx,$dy"}) {
	# We need to calculate a route to this destination
	CalcRoute($dx,$dy);
    }
    $a=substr(${$route{"$dx,$dy"}}[$cy],$cx-1,1);
    if ($a eq 'S' || $a eq 'N' || $a eq 'W' || $a eq 'E') {
	return $a;
    } else {
	return 0;
    }
}
